package info.textgrid.middleware;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.action.search.SearchScrollRequest;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import info.textgrid.middleware.oaipmh.RequestType;
import info.textgrid.middleware.oaipmh.ResumptionTokenType;

/**
 *
 */
public abstract class RecordListDelivererAbstract implements RecordListDelivererInterface {

  protected OAI_ESClient oaiEsClient;

  protected boolean textgrid;
  protected boolean dariah;

  protected String[] fields;
  protected String[] workFields;

  protected String dateOfObjectCreation;
  protected String relationToFurtherMetadataObject;
  protected String repositoryObjectURIPrefix;
  protected String modifiedField;
  protected String identifierField;
  protected String rangeField;
  protected String formatField;
  protected String formatToFilter;
  protected int searchResponseSize;
  protected String specField;
  protected String specFieldPrefix;

  protected String modifiedValue;
  protected int theVerySpecialIDIOMCursorWidth;
  private long resultSize;
  private boolean foundItems;

  public ResumptionTokenType resTokenForResponse;

  // private static final int LIFETIME_RES_TOKEN = 600;
  protected static Map<String, Integer> cursorCollector = new Hashtable<String, Integer>();

  private static Log log = LogFactory.getLog(RecordListDelivererAbstract.class);

  /**
   * @param textgrid
   * @param dariah
   */
  public RecordListDelivererAbstract(boolean textgrid, boolean dariah) {
    this.textgrid = textgrid;
    this.dariah = dariah;
  }

  /**
   * @param query
   * @param resumptionToken
   * @param set
   * @return
   */
  protected List<String> getFieldsFromESIndex(QueryBuilder query, String resumptionToken,
      String set) {

    List<String> uriList = new ArrayList<String>();

    QueryBuilder recordFilter;
    if (this.textgrid) {
      // We filter out all editions here!
      recordFilter = QueryBuilders.boolQuery().must(query)
          .must(QueryBuilders.matchPhraseQuery("format", this.formatToFilter));
    } else {
      // Do not filter at all in DH. We need every ID!
      recordFilter = QueryBuilders.boolQuery().must(query);
    }

    SearchRequest searchRequest = new SearchRequest(OAI_ESClient.getEsIndex());
    SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder();

    searchSourceBuilder.query(recordFilter);
    searchSourceBuilder.size(this.theVerySpecialIDIOMCursorWidth);
    searchRequest.source(searchSourceBuilder);

    SearchResponse scrollResp = new SearchResponse();

    if (resumptionToken != null) {
      SearchScrollRequest scrollRequest = new SearchScrollRequest(resumptionToken);
      scrollRequest.scroll(TimeValue.timeValueHours(24L));

      try {
        scrollResp = OAI_ESClient.getEsClient().scroll(scrollRequest, RequestOptions.DEFAULT);
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    } else {
      searchRequest.source(searchSourceBuilder);
      searchRequest.scroll(TimeValue.timeValueHours(24L));
      try {
        scrollResp = OAI_ESClient.getEsClient().search(searchRequest, RequestOptions.DEFAULT);
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    }

    String scrollID = scrollResp.getScrollId();
    long completeListSize = scrollResp.getHits().totalHits;
    setResultSize(completeListSize);

    if (completeListSize > 0) {

      setFoundItems(true);
      int i = 0;

      for (SearchHit hit : scrollResp.getHits().getHits()) {
        i++;
        if (hit != null && hit.getFields() != null) {
          String id2add;
          // FIXME Could we not use hit.getId() also for TG hits? Where is the difference?
          if (this.textgrid) {
            id2add = hit.getSourceAsMap().get(TGConstants.URI).toString();
          } else {
            id2add = hit.getId();
          }
          uriList.add(id2add);
        }
      }
      if (resumptionToken != null
          && this.resTokenForResponse.getCursor().intValue() >= completeListSize) {
        try {
          cursorCollector.remove(resumptionToken);
        } catch (NullPointerException couldNotRemove) {
          log.info("Could not remove hash value: " + resumptionToken + " from hash map");
        }
        this.resTokenForResponse.setValue("");
      } else {
        this.resTokenForResponse = OAIPMHUtilities.getResumptionToken(
            completeListSize, resumptionToken, cursorCollector, scrollID, this.searchResponseSize,
            i);
      }

      log.debug("cursorCollector: " + cursorCollector);

    } else {
      log.info("HALLO");
      setFoundItems(false);
    }

    if (this.resTokenForResponse != null) {
      log.debug("restok value: " + this.resTokenForResponse.getValue());
      log.debug("restok cursor: " + this.resTokenForResponse.getCursor());
    }

    return uriList;
  }

  /**
   * @param request
   * @return
   */
  public static ErrorHandler requestChecker(RequestType request) {

    ErrorHandler result = new ErrorHandler();

    // Check if metadata prefix is existing and valid.
    if (request.getMetadataPrefix() != null
        && !request.getMetadataPrefix().equals(OAIPMHUtilities.OAIDC_PREFIX)
        && !request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_IDIOM_PREFIX)
        && !request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_OPENAIRE_PREFIX)) {
      result.setError(OAIPMHConstants.OAI_METADATA_FORMAT_ERROR,
          "The value of the metadataPrefix: " + request.getMetadataPrefix()
              + " is not supported by the item identified by the value of: "
              + request.getIdentifier());
    }

    // Check if resumptionToken is invalid or existing.
    if (request.getResumptionToken() != null) {
      boolean restokIDIOMExisting = RecordListDelivererIDIOM.cursorCollector != null
          && RecordListDelivererIDIOM.cursorCollector.containsKey(request.getResumptionToken());

      boolean restokDCExisting = RecordListDelivererDC.cursorCollector != null
          && RecordListDelivererDC.cursorCollector.containsKey(request.getResumptionToken());
      boolean restok = cursorCollector != null
          && cursorCollector.containsKey(request.getResumptionToken());

      if (!restokDCExisting && !restokIDIOMExisting && !restok) {
        result.setError(OAIPMHConstants.OAI_BAD_RESUMPTION_TOKEN, "The value of the "
            + request.getResumptionToken() + " argument is invalid or expired.");
      }
    }

    // Check params in general.
    List<String> errorValues = new ArrayList<String>();

    if (request.getResumptionToken() == null && request.getMetadataPrefix() == null) {
      errorValues.add("metadataPrefix");
    }
    if (request.getIdentifier() != null) {
      errorValues.add("identifier");
    }
    /*
     * if(request.getFrom()!= null && (!OAIPMHUtilities.isThisDateValid(request.getFrom()) &&
     * !OAIPMHUtilities.isThisDateValidToOtherTimeStamp(request.getFrom()))) {
     * errorValues.add("from"); } if(request.getUntil() != null &&
     * (!OAIPMHUtilities.isThisDateValid(request.getUntil()) &&
     * !OAIPMHUtilities.isThisDateValidToOtherTimeStamp(request.getUntil()))) {
     * errorValues.add("until"); } if(request.getFrom()!= null && request.getUntil()!=null &&
     * !OAIPMHUtilities.getFormatOfDate(request.getFrom()).equals(OAIPMHUtilities.getFormatOfDate(
     * request.getFrom()))) { errorValues.add("until, from"); }
     */
    if (errorValues.size() > 0) {
      result.setError(OAIPMHConstants.OAI_BAD_ARGUMENT, "The request includes illegal arguments "
          + "or is missing required arguments: " + errorValues);
    }

    return result;
  }

  // /*
  // * (non-Javadoc)
  // *
  // * @see info.textgrid.middleware.RecordListDelivererInterface#setHeader(java.lang.String,
  // * java.lang.String, java.lang.String)
  // */
  // public HeaderType setHeader(final String set, final String headerIdentifier,
  // String modifiedDate) {
  //
  // HeaderType header = new HeaderType();
  // String identifierForHeader = headerIdentifier;
  //
  // System.out.println("header identifier: " + headerIdentifier);
  // System.out.println("modified value: " + modifiedDate);
  //
  // // Set date in XML format.
  // try {
  // header.setDatestamp(OAIPMHUtilities.convertDateFormat(modifiedDate).toXMLFormat());
  // } catch (ParseException e) {
  // // TODO Auto-generated catch block
  // e.printStackTrace();
  // } catch (DatatypeConfigurationException e) {
  // // TODO Auto-generated catch block
  // e.printStackTrace();
  // }
  //
  // // Set header specific for TextGrid objects (remove URL).
  // if (this.textgrid == true) {
  // identifierForHeader = identifierForHeader.replaceFirst("https://textgridrep.org/", "");
  // }
  //
  // header.setIdentifier(identifierForHeader);
  //
  // // Set set :-)
  // if (set != null) {
  // header.getSetSpec().add(this.specFieldPrefix + set);
  // }
  //
  // return header;
  // }

  // **
  // GETTER AND SETTER
  // **

  /**
   * @return
   */
  public String getRangeField() {
    return this.rangeField;
  }

  /**
   * @param rangeField
   */
  public void setRangeField(String rangeField) {
    this.rangeField = rangeField;
  }

  /**
   * @param formatToFilter
   */
  public void setFormatToFilter(String formatToFilter) {
    this.formatToFilter = formatToFilter;
  }

  /**
   * @param formatField
   */
  public void setFormatField(String formatField) {
    this.formatField = formatField;
  }

  /**
   * @return
   */
  public int getSearchResponseSize() {
    return this.searchResponseSize;
  }

  /**
   * @param searchResponseSize
   */
  public void setSearchResponseSize(int searchResponseSize) {
    this.searchResponseSize = searchResponseSize;
  }

  /**
   * @return
   */
  public long getResultSize() {
    return this.resultSize;
  }

  /**
   * @param resultSize
   */
  public void setResultSize(long resultSize) {
    this.resultSize = resultSize;
  }

  /**
   * @return
   */
  public boolean isFoundItems() {
    return this.foundItems;
  }

  /**
   * @param foundItems
   */
  public void setFoundItems(boolean foundItems) {
    this.foundItems = foundItems;
  }

  /**
   * @return
   */
  public OAI_ESClient getOaiEsClient() {
    return this.oaiEsClient;
  }

  /**
   * @param oaiEsClient
   */
  public void setOaiEsClient(OAI_ESClient oaiEsClient) {
    this.oaiEsClient = oaiEsClient;
  }

  /**
   * @return
   */
  public ResumptionTokenType getResTokenForResponse() {
    return this.resTokenForResponse;
  }

  /**
   * @param resTokenForResponse
   */
  public void setResTokenForResponse(ResumptionTokenType resTokenForResponse) {
    this.resTokenForResponse = resTokenForResponse;
  }


  /**
   * @return
   */
  public String getModifiedField() {
    return this.modifiedField;
  }

  /**
   * @param modifiedField
   */
  public void setModifiedField(String modifiedField) {
    this.modifiedField = modifiedField;
  }


  /**
   * @return
   */
  public String getDateOfObjectCreation() {
    return this.dateOfObjectCreation;
  }

  /**
   * @param dateOfObjectCreation
   */
  public void setDateOfObjectCreation(String dateOfObjectCreation) {
    this.dateOfObjectCreation = dateOfObjectCreation;
  }

  /**
   * @return
   */
  public String getRelationToFurtherMetadataObject() {
    return this.relationToFurtherMetadataObject;
  }

  /**
   * @param relationToFurtherMetadataObject
   */
  public void setRelationToFurtherMetadataObject(String relationToFurtherMetadataObject) {
    this.relationToFurtherMetadataObject = relationToFurtherMetadataObject;
  }

  /**
   * @return
   */
  public String getRepositoryObjectURIPrefix() {
    return this.repositoryObjectURIPrefix;
  }

  /**
   * @param repositoryObjectURIPrefix
   */
  public void setRepositoryObjectURIPrefix(String repositoryObjectURIPrefix) {
    this.repositoryObjectURIPrefix = repositoryObjectURIPrefix;
  }

  /**
   * @param textgrid
   */
  public void setTextgrid(boolean textgrid) {
    this.textgrid = textgrid;
  }

  /**
   * @return
   */
  public boolean isDariah() {
    return this.dariah;
  }

  /**
   * @param dariah
   */
  public void setDariah(boolean dariah) {
    this.dariah = dariah;
  }

  /**
   * @return
   */
  public String getIdentifierField() {
    return this.identifierField;
  }

  /**
   * @param identifierField
   */
  public void setIdentifierField(String identifierField) {
    this.identifierField = identifierField;
  }

  /**
   * @return
   */
  public String getModifiedValue() {
    return this.modifiedValue;
  }

  /**
   * @param modifiedValue
   */
  public void setModifiedValue(String modifiedValue) {
    this.modifiedValue = modifiedValue;
  }

  /**
   * @return
   */
  public String getFormatField() {
    return this.formatField;
  }

  /**
   * @return
   */
  public String getFormatToFilter() {
    return this.formatToFilter;
  }


  /**
   * @return
   */
  public String[] getWorkFields() {
    return this.workFields;
  }

  /**
   * @param workFields
   */
  public void setWorkFields(String[] workFields) {
    this.workFields = workFields;
  }

  /**
   * @param fields
   */
  public void setFields(String[] fields) {
    this.fields = fields;
  }

  /**
   * @return
   */
  public String[] getFields() {
    return this.fields;
  }

  /**
   * @return
   */
  public String getSpecField() {
    return this.specField;
  }

  /**
   * @param specField
   */
  public void setSpecField(String specField) {
    this.specField = specField;
  }

  /**
   * @return
   */
  public String getSpecFieldPrefix() {
    return this.specFieldPrefix;
  }

  /**
   * @param specFieldPrefix
   */
  public void setSpecFieldPrefix(String specFieldPrefix) {
    this.specFieldPrefix = specFieldPrefix;
  }

  /**
   * @return
   */
  public int getTheVerySpecialIDIOMCursorWidth() {
    return this.theVerySpecialIDIOMCursorWidth;
  }

  /**
   * @param theVerySpecialIDIOMCursorWidth
   */
  public void setTheVerySpecialIDIOMCursorWidth(int theVerySpecialIDIOMCursorWidth) {
    this.theVerySpecialIDIOMCursorWidth = theVerySpecialIDIOMCursorWidth;
  }

}
